@@ -37,10 +37,11 @@ from collections.abc import (
     Iterable,
     MutableMapping,
 )
+from functools import partial
 
 from PyQt5.QtCore import QObject
 
-from picard import config
+from picard.config import get_config
 from picard.mbjson import (
     artist_credit_from_node,
     get_score,
@@ -174,9 +175,8 @@ class Metadata(MutableMapping):
         self.images = ImageList()
         self.has_common_images = True
 
-        d = dict(*args, **kwargs)
-        for k, v in d.items():
-            self[k] = v
+        if args or kwargs:
+            self.update(*args, **kwargs)
         if images is not None:
             for image in images:
                 self.images.append(image)
@@ -292,6 +292,7 @@ class Metadata(MutableMapping):
 
         parts.append((date_match_factor, weights['date']))
 
+        config = get_config()
         weights_from_preferred_countries(parts, release,
                                          config.setting["preferred_release_countries"],
                                          weights["releasecountry"])
@@ -361,7 +362,7 @@ class Metadata(MutableMapping):
 
     def update(self, *args, **kwargs):
         one_arg = len(args) == 1
-        if one_arg and isinstance(args[0], self.__class__):
+        if one_arg and (isinstance(args[0], self.__class__) or isinstance(args[0], MultiMetadataProxy)):
             self._update_from_metadata(args[0])
         elif one_arg and isinstance(args[0], MutableMapping):
             # update from MutableMapping (ie. dict)
@@ -375,6 +376,16 @@ class Metadata(MutableMapping):
             # no argument, raise TypeError to mimic dict.update()
             raise TypeError("descriptor 'update' of '%s' object needs an argument" % self.__class__.__name__)
 
+    def diff(self, other):
+        """Returns a new Metadata object with only the tags that changed in self compared to other"""
+        m = Metadata()
+        for tag, values in self.rawitems():
+            other_values = other.getall(tag)
+            if other_values != values:
+                m[tag] = values
+        m.deleted_tags = self.deleted_tags - other.deleted_tags
+        return m
+
     def _update_from_metadata(self, other, copy_images=True):
         for k, v in other.rawitems():
             self.set(k, v[:])
@@ -510,6 +521,90 @@ class Metadata(MutableMapping):
         return ("store: %r\ndeleted: %r\nimages: %r\nlength: %r" % (self._store, self.deleted_tags, [str(img) for img in self.images], self.length))
 
 
+class MultiMetadataProxy:
+    """
+    Wraps a writable Metadata object together with another
+    readonly Metadata object.
+
+    Changes are written to the writable object, while values are
+    read from both the writable and the readonly object (with the writable
+    object taking precedence). The use case is to provide access to Metadata
+    values without making them part of the actual Metadata. E.g. allow track
+    metadata to use file specific metadata, without making it actually part
+    of the track.
+    """
+
+    WRITE_METHODS = [
+        'add_unique',
+        'add',
+        'apply_func',
+        'clear_deleted',
+        'clear',
+        'copy',
+        'delete',
+        'pop',
+        'set',
+        'strip_whitespace',
+        'unset',
+        'update',
+    ]
+
+    def __init__(self, metadata, *readonly_metadata):
+        self.metadata = metadata
+        self.combined_metadata = Metadata()
+        for m in reversed(readonly_metadata):
+            self.combined_metadata.update(m)
+        self.combined_metadata.update(metadata)
+
+    def __getattr__(self, name):
+        if name in self.WRITE_METHODS:
+            return partial(self.__write, name)
+        else:
+            attribute = self.combined_metadata.__getattribute__(name)
+            if callable(attribute):
+                return partial(self.__read, name)
+            else:
+                return attribute
+
+    def __setattr__(self, name, value):
+        if name in ('metadata', 'combined_metadata'):
+            super().__setattr__(name, value)
+        else:
+            self.metadata.__setattr__(name, value)
+            self.combined_metadata.__setattr__(name, value)
+
+    def __write(self, name, *args, **kwargs):
+        func1 = self.metadata.__getattribute__(name)
+        func2 = self.combined_metadata.__getattribute__(name)
+        func1(*args, **kwargs)
+        return func2(*args, **kwargs)
+
+    def __read(self, name, *args, **kwargs):
+        func = self.combined_metadata.__getattribute__(name)
+        return func(*args, **kwargs)
+
+    def __getitem__(self, name):
+        return self.__read('__getitem__', name)
+
+    def __setitem__(self, name, values):
+        return self.__write('__setitem__', name, values)
+
+    def __delitem__(self, name):
+        return self.__write('__delitem__', name)
+
+    def __iter__(self):
+        return self.__read('__iter__')
+
+    def __len__(self):
+        return self.__read('__len__')
+
+    def __contains__(self, name):
+        return self.__read('__contains__', name)
+
+    def __repr__(self):
+        return self.__read('__repr__')
+
+
 _album_metadata_processors = PluginFunctions(label='album_metadata_processors')
 _track_metadata_processors = PluginFunctions(label='track_metadata_processors')
 
