@@ -10,7 +10,7 @@
 # Copyright (C) 2013-2014, 2017-2019 Laurent Monin
 # Copyright (C) 2015 Ohm Patel
 # Copyright (C) 2015 Wieland Hoffmann
-# Copyright (C) 2015, 2018-2020 Philipp Wolfer
+# Copyright (C) 2015, 2018-2021 Philipp Wolfer
 # Copyright (C) 2016-2018 Sambhav Kothari
 # Copyright (C) 2018 Vishal Choudhary
 # Copyright (C) 2020 Gabriel Ferreira
@@ -39,16 +39,20 @@ from PyQt5 import (
     QtWidgets,
 )
 
-from picard import config
 from picard.album import Album
-from picard.browser.browser import BrowserIntegration
 from picard.browser.filelookup import FileLookup
 from picard.cluster import Cluster
+from picard.config import (
+    BoolOption,
+    Option,
+    get_config,
+)
 from picard.file import File
 from picard.metadata import MULTI_VALUED_JOINER
 from picard.track import Track
 from picard.util import (
     format_time,
+    icontheme,
     restore_method,
     thread,
     throttle,
@@ -71,6 +75,7 @@ class TagStatus:
     CHANGED = ADDED | REMOVED
     EMPTY = 8
     NOTREMOVABLE = 16
+    READONLY = 32
 
 
 class TagCounter(dict):
@@ -130,13 +135,16 @@ class TagDiff(object):
         else:
             return orig != new
 
-    def add(self, tag, orig_values, new_values, removable, removed=False):
+    def add(self, tag, orig_values, new_values, removable, removed=False, readonly=False, top_tags=None):
         if orig_values:
             self.orig.add(tag, orig_values)
 
         if new_values:
             self.new.add(tag, new_values)
 
+        if not top_tags:
+            top_tags = set()
+
         if (orig_values and not new_values) or removed:
             self.status[tag] |= TagStatus.REMOVED
         elif new_values and not orig_values:
@@ -144,7 +152,7 @@ class TagDiff(object):
             removable = True
         elif orig_values and new_values and self.__tag_ne(tag, orig_values, new_values):
             self.status[tag] |= TagStatus.CHANGED
-        elif not (orig_values or new_values or tag in config.setting['metadatabox_top_tags']):
+        elif not (orig_values or new_values or tag in top_tags):
             self.status[tag] |= TagStatus.EMPTY
         else:
             self.status[tag] |= TagStatus.NOCHANGE
@@ -152,6 +160,9 @@ class TagDiff(object):
         if not removable:
             self.status[tag] |= TagStatus.NOTREMOVABLE
 
+        if readonly:
+            self.status[tag] |= TagStatus.READONLY
+
     def tag_status(self, tag):
         status = self.status[tag]
         for s in (TagStatus.CHANGED, TagStatus.ADDED,
@@ -188,12 +199,16 @@ class TableTagEditorDelegate(TagEditorDelegate):
 class MetadataBox(QtWidgets.QTableWidget):
 
     options = (
-        config.Option("persist", "metadatabox_header_state", QtCore.QByteArray()),
-        config.BoolOption("persist", "show_changes_first", False)
+        Option("persist", "metadatabox_header_state", QtCore.QByteArray()),
+        BoolOption("persist", "show_changes_first", False)
     )
 
+    COLUMN_ORIG = 1
+    COLUMN_NEW = 2
+
     def __init__(self, parent):
         super().__init__(parent)
+        config = get_config()
         self.parent = parent
         self.setAccessibleName(_("metadata view"))
         self.setAccessibleDescription(_("Displays original and new tags for the selected files"))
@@ -224,7 +239,6 @@ class MetadataBox(QtWidgets.QTableWidget):
         self.changes_first_action.setCheckable(True)
         self.changes_first_action.setChecked(config.persist["show_changes_first"])
         self.changes_first_action.toggled.connect(self.toggle_changes_first)
-        self.browser_integration = BrowserIntegration()
         # TR: Keyboard shortcut for "Add New Tag..."
         self.add_tag_shortcut = QtWidgets.QShortcut(QtGui.QKeySequence(_("Alt+Shift+A")), self, partial(self.edit_tag, ""))
         self.add_tag_action.setShortcut(self.add_tag_shortcut.key())
@@ -235,12 +249,14 @@ class MetadataBox(QtWidgets.QTableWidget):
         self.preserved_tags = PreservedTags()
         self._single_file_album = False
         self._single_track_album = False
+        self.tagger.clipboard().dataChanged.connect(self.update_clipboard)
 
     def get_file_lookup(self):
         """Return a FileLookup object."""
+        config = get_config()
         return FileLookup(self, config.setting["server_host"],
                           config.setting["server_port"],
-                          self.browser_integration.port)
+                          self.tagger.browser_integration.port)
 
     def lookup_tags(self):
         lookup = self.get_file_lookup()
@@ -264,7 +280,7 @@ class MetadataBox(QtWidgets.QTableWidget):
             lookup_func(v)
 
     def edit(self, index, trigger, event):
-        if index.column() != 2:
+        if index.column() != self.COLUMN_NEW:
             return False
         item = self.itemFromIndex(index)
         if item.flags() & QtCore.Qt.ItemIsEditable and \
@@ -282,20 +298,43 @@ class MetadataBox(QtWidgets.QTableWidget):
                 return super().edit(index, trigger, event)
         return False
 
-    def event(self, e):
+    def keyPressEvent(self, event):
+        if event.matches(QtGui.QKeySequence.Copy):
+            self.copy_value()
+        elif event.matches(QtGui.QKeySequence.Paste):
+            self.paste_value()
+        else:
+            super().keyPressEvent(event)
+
+    def copy_value(self):
         item = self.currentItem()
-        if (item and e.type() == QtCore.QEvent.KeyPress and e.modifiers() == QtCore.Qt.ControlModifier):
+        if item:
             column = item.column()
             tag = self.tag_diff.tag_names[item.row()]
-            if e.key() == QtCore.Qt.Key_C:
-                if column == 1:
-                    self.clipboard = list(self.tag_diff.orig[tag])
-                elif column == 2:
-                    self.clipboard = list(self.tag_diff.new[tag])
-            elif e.key() == QtCore.Qt.Key_V and column == 2 and tag != "~length":
-                self.set_tag_values(tag, list(self.clipboard))
+            value = None
+            if column == self.COLUMN_ORIG:
+                value = self.tag_diff.orig[tag]
+            elif column == self.COLUMN_NEW:
+                value = self.tag_diff.new[tag]
+            if tag == '~length':
+                value = [format_time(value or 0), ]
+            if value is not None:
+                self.tagger.clipboard().setText(MULTI_VALUED_JOINER.join(value))
+                self.clipboard = value
+
+    def paste_value(self):
+        item = self.currentItem()
+        if item:
+            column = item.column()
+            tag = self.tag_diff.tag_names[item.row()]
+            if column == self.COLUMN_NEW and self.tag_is_editable(tag):
+                self.set_tag_values(tag, self.clipboard)
                 self.update()
-        return super().event(e)
+
+    def update_clipboard(self):
+        clipboard = self.tagger.clipboard().text().split(MULTI_VALUED_JOINER)
+        if clipboard:
+            self.clipboard = clipboard
 
     def closeEditor(self, editor, hint):
         super().closeEditor(editor, hint)
@@ -320,20 +359,25 @@ class MetadataBox(QtWidgets.QTableWidget):
     def contextMenuEvent(self, event):
         menu = QtWidgets.QMenu(self)
         if self.objects:
-            tags = self.selected_tags(discard=('~length',))
-            if len(tags) == 1:
+            tags = self.selected_tags()
+            single_tag = len(tags) == 1
+            if single_tag:
                 selected_tag = tags[0]
+                editable = self.tag_is_editable(selected_tag)
                 edit_tag_action = QtWidgets.QAction(_("Edit..."), self.parent)
                 edit_tag_action.triggered.connect(partial(self.edit_tag, selected_tag))
                 edit_tag_action.setShortcut(self.edit_tag_shortcut.key())
+                edit_tag_action.setEnabled(editable)
                 menu.addAction(edit_tag_action)
                 if selected_tag not in self.preserved_tags:
                     add_to_preserved_tags_action = QtWidgets.QAction(_("Add to 'Preserve Tags' List"), self.parent)
                     add_to_preserved_tags_action.triggered.connect(partial(self.preserved_tags.add, selected_tag))
+                    add_to_preserved_tags_action.setEnabled(editable)
                     menu.addAction(add_to_preserved_tags_action)
                 else:
                     remove_from_preserved_tags_action = QtWidgets.QAction(_("Remove from 'Preserve Tags' List"), self.parent)
                     remove_from_preserved_tags_action.triggered.connect(partial(self.preserved_tags.discard, selected_tag))
+                    remove_from_preserved_tags_action.setEnabled(editable)
                     menu.addAction(remove_from_preserved_tags_action)
             removals = []
             useorigs = []
@@ -342,8 +386,8 @@ class MetadataBox(QtWidgets.QTableWidget):
                 column = item.column()
                 for tag in tags:
                     if tag in self.lookup_tags().keys():
-                        if (column == 1 or column == 2) and len(tags) == 1 and item.text():
-                            if column == 1:
+                        if (column == self.COLUMN_ORIG or column == self.COLUMN_NEW) and single_tag and item.text():
+                            if column == self.COLUMN_ORIG:
                                 values = self.tag_diff.orig[tag]
                             else:
                                 values = self.tag_diff.new[tag]
@@ -358,7 +402,7 @@ class MetadataBox(QtWidgets.QTableWidget):
                         track_albums = set()
                         for file in self.files:
                             objects = [file]
-                            if file.parent in self.tracks and len(self.files & set(file.parent.linked_files)) == 1:
+                            if file.parent in self.tracks and len(self.files & set(file.parent.files)) == 1:
                                 objects.append(file.parent)
                                 file_tracks.append(file.parent)
                                 track_albums.add(file.parent.album)
@@ -373,18 +417,29 @@ class MetadataBox(QtWidgets.QTableWidget):
                             objects = [album]
                             orig_values = list(album.orig_metadata.getall(tag)) or [""]
                             useorigs.append(partial(self.set_tag_values, tag, orig_values, objects))
-                if removals:
-                    remove_tag_action = QtWidgets.QAction(_("Remove"), self.parent)
-                    remove_tag_action.triggered.connect(partial(self._apply_update_funcs, removals))
-                    remove_tag_action.setShortcut(self.remove_tag_shortcut.key())
-                    menu.addAction(remove_tag_action)
+                remove_tag_action = QtWidgets.QAction(_("Remove"), self.parent)
+                remove_tag_action.triggered.connect(partial(self._apply_update_funcs, removals))
+                remove_tag_action.setShortcut(self.remove_tag_shortcut.key())
+                remove_tag_action.setEnabled(bool(removals))
+                menu.addAction(remove_tag_action)
                 if useorigs:
                     name = ngettext("Use Original Value", "Use Original Values", len(useorigs))
                     use_orig_value_action = QtWidgets.QAction(name, self.parent)
                     use_orig_value_action.triggered.connect(partial(self._apply_update_funcs, useorigs))
                     menu.addAction(use_orig_value_action)
                     menu.addSeparator()
-            if len(tags) == 1 or removals or useorigs:
+                if single_tag:
+                    menu.addSeparator()
+                    copy_action = QtWidgets.QAction(icontheme.lookup('edit-copy', icontheme.ICON_SIZE_MENU), _("&Copy"), self)
+                    copy_action.triggered.connect(self.copy_value)
+                    copy_action.setShortcut(QtGui.QKeySequence.Copy)
+                    menu.addAction(copy_action)
+                    paste_action = QtWidgets.QAction(icontheme.lookup('edit-paste', icontheme.ICON_SIZE_MENU), _("&Paste"), self)
+                    paste_action.triggered.connect(self.paste_value)
+                    paste_action.setShortcut(QtGui.QKeySequence.Paste)
+                    paste_action.setEnabled(editable)
+                    menu.addAction(paste_action)
+            if single_tag or removals or useorigs:
                 menu.addSeparator()
             menu.addAction(self.add_tag_action)
             menu.addSeparator()
@@ -401,35 +456,35 @@ class MetadataBox(QtWidgets.QTableWidget):
         EditTagDialog(self.parent, tag).exec_()
 
     def edit_selected_tag(self):
-        tags = self.selected_tags(discard=('~length',))
+        tags = self.selected_tags(filter_func=self.tag_is_editable)
         if len(tags) == 1:
             self.edit_tag(tags[0])
 
     def toggle_changes_first(self, checked):
+        config = get_config()
         config.persist["show_changes_first"] = checked
         self.update()
 
     def set_tag_values(self, tag, values, objects=None):
         if objects is None:
             objects = self.objects
-        self.parent.ignore_selection_changes = True
-        if values == [""]:
-            values = []
-        if not values and self.tag_is_removable(tag):
-            for obj in objects:
-                del obj.metadata[tag]
-                obj.update()
-        elif values:
-            for obj in objects:
-                obj.metadata[tag] = values
-                obj.update()
-        self.parent.ignore_selection_changes = False
+        with self.parent.ignore_selection_changes:
+            if values == [""]:
+                values = []
+            if not values and self.tag_is_removable(tag):
+                for obj in objects:
+                    del obj.metadata[tag]
+                    obj.update()
+            elif values:
+                for obj in objects:
+                    obj.metadata[tag] = values
+                    obj.update()
 
     def remove_tag(self, tag):
         self.set_tag_values(tag, [])
 
     def remove_selected_tags(self):
-        for tag in self.selected_tags(discard=('~length',)):
+        for tag in self.selected_tags(filter_func=self.tag_is_removable):
             if self.tag_is_removable(tag):
                 self.remove_tag(tag)
         self.parent.update_selection(new_selection=False, drop_album_caches=True)
@@ -437,12 +492,15 @@ class MetadataBox(QtWidgets.QTableWidget):
     def tag_is_removable(self, tag):
         return self.tag_diff.status[tag] & TagStatus.NOTREMOVABLE == 0
 
-    def selected_tags(self, discard=None):
-        if discard is None:
-            discard = set()
+    def tag_is_editable(self, tag):
+        return self.tag_diff.status[tag] & TagStatus.READONLY == 0
+
+    def selected_tags(self, filter_func=None):
         tags = set(self.tag_diff.tag_names[item.row()]
                    for item in self.selectedItems())
-        return list(tags.difference(discard))
+        if filter_func:
+            tags = filter(filter_func, tags)
+        return list(tags)
 
     def _update_selection(self):
         files = set()
@@ -453,7 +511,7 @@ class MetadataBox(QtWidgets.QTableWidget):
                 files.add(obj)
             elif isinstance(obj, Track):
                 tracks.add(obj)
-                files.update(obj.linked_files)
+                files.update(obj.files)
             elif isinstance(obj, Cluster) and obj.can_edit_tags():
                 objects.add(obj)
                 files.update(obj.files)
@@ -461,7 +519,7 @@ class MetadataBox(QtWidgets.QTableWidget):
                 objects.add(obj)
                 tracks.update(obj.tracks)
                 for track in obj.tracks:
-                    files.update(track.linked_files)
+                    files.update(track.files)
         objects.update(files)
         objects.update(tracks)
         self.selection_dirty = False
@@ -516,15 +574,15 @@ class MetadataBox(QtWidgets.QTableWidget):
             TagStatus.CHANGED: QtGui.QBrush(interface_colors.get_qcolor('tagstatus_changed'))
         }
 
+        config = get_config()
         tag_diff = TagDiff(max_length_diff=config.setting["ignore_track_duration_difference_under"])
         orig_tags = tag_diff.orig
         new_tags = tag_diff.new
-        # existing_tags are orig_tags that would not be overwritten by
-        # any new_tags, assuming clear_existing_tags is disabled.
-        existing_tags = set()
         tag_diff.objects = len(files)
 
         clear_existing_tags = config.setting["clear_existing_tags"]
+        top_tags = config.setting['metadatabox_top_tags']
+        top_tags_set = set(top_tags)
 
         for file in files:
             new_metadata = file.new_metadata
@@ -535,32 +593,32 @@ class MetadataBox(QtWidgets.QTableWidget):
                 new_values = new_metadata.getall(name)
                 orig_values = orig_metadata.getall(name)
 
-                if not ((new_values and name not in existing_tags) or clear_existing_tags):
+                if not clear_existing_tags and not new_values:
                     new_values = list(orig_values or [""])
-                    existing_tags.add(name)
 
                 removed = name in new_metadata.deleted_tags
-                tag_diff.add(name, orig_values, new_values, True, removed)
+                tag_diff.add(name, orig_values, new_values, True, removed, top_tags=top_tags_set)
 
-            tag_diff.add("~length",
-                         str(orig_metadata.length), str(new_metadata.length), False)
+            tag_diff.add("~length", str(orig_metadata.length), str(new_metadata.length),
+                         removable=False, readonly=True)
 
         for track in tracks:
             if track.num_linked_files == 0:
-                for name, values in track.metadata.rawitems():
+                for name, new_values in track.metadata.rawitems():
                     if not name.startswith("~"):
                         if name in track.orig_metadata:
-                            tag_diff.add(name, [track.orig_metadata[name]], values, True)
+                            orig_values = track.orig_metadata.getall(name)
                         else:
-                            tag_diff.add(name, values, values, True)
+                            orig_values = new_values
+                        tag_diff.add(name, orig_values, new_values, True)
 
                 length = str(track.metadata.length)
-                tag_diff.add("~length", length, length, False)
+                tag_diff.add("~length", length, length, removable=False, readonly=True)
 
                 tag_diff.objects += 1
 
         all_tags = set(list(orig_tags.keys()) + list(new_tags.keys()))
-        common_tags = [tag for tag in config.setting['metadatabox_top_tags'] if tag in all_tags]
+        common_tags = [tag for tag in top_tags if tag in all_tags]
         tag_names = common_tags + sorted(all_tags.difference(common_tags),
                                          key=lambda x: display_tag_name(x).lower())
 
@@ -655,12 +713,14 @@ class MetadataBox(QtWidgets.QTableWidget):
 
     @restore_method
     def restore_state(self):
+        config = get_config()
         state = config.persist["metadatabox_header_state"]
         header = self.horizontalHeader()
         header.restoreState(state)
         header.setSectionResizeMode(QtWidgets.QHeaderView.Interactive)
 
     def save_state(self):
+        config = get_config()
         header = self.horizontalHeader()
         state = header.saveState()
         config.persist["metadatabox_header_state"] = state
